-- Trivial eiffel pretty printer 
-- public domain FA 1996-05

indexing
	description: "Trivial Eiffel pretty printer";
	licence: public_domain;
	interface: inheritance;

deferred class EIFFEL_PRINT

feature 
	
	make is
		do
		end
				
feature { NONE } -- Keywords (should be cleverer)

	is_keyword(str : STRING) : BOOLEAN is
		local
			i,up : INTEGER;
			lowkey : STRING
		do
			lowkey := clone(str)
			lowkey.to_lower

			from
				i := keywords.lower
				up := keywords.upper
			variant
				up + 1 - i
			until
				Result = true or i > up
			loop
				if lowkey.is_equal(keywords.item(i)) then
					Result := true
				end
				i := i + 1
			end
		end
	
	keywords : ARRAY[STRING] is 
		once
			Result := << 
				"alias",	"all",		"and",		"as",		"check",	"class",
				"creation",	"Current",	"debug",	"deferred",	"do",
				"else",		"elseif",	"end",		"ensure",	"expanded",	"export",
				"external",	"false",	"feature",	"from",		"frozen",	"if",
				"implies",	"indexing",	"infix",	"inherit",	"inspect",	"invariant",
				"is",		"like",		"local",	"loop",		"not",		"obsolete",
				"old",		"once",		"or",		"prefix",	"redefine",	"rename",
				"require",	"rescue",	"Result",	"retry",	"select",	"separate",
				"strip",	"then",		"true",		"undefine",	"unique",	"until",
				"variant",	"Void",		"when",		"xor", 		"Precursor" >>
		end

feature	-- Public interface
	
	print_eiffel_line(str : STRING) is
			-- Process one line of source code.
		require
			ok: str /= Void
		do
			current_string := str
			process_line
			print_new_line
		end

	print_prologue is
			-- Print prologue before output.
		deferred
		end

	print_epilogue is
			-- Print epilogue after output.
		deferred
		end

feature -- Verbatim

	print_verbatim(str : STRING) is
			-- Print this string going only through character filter, 
			-- useful for actual printing in effective filters when 
			-- some characters have to be escaped.
		require
			ok: str /= Void
		local
			i : INTEGER
		do	
			from i := 1 variant str.count + 1 - i until i > str.count loop
				print_character(str.item(i))
				i := i + 1			
			end
		end

feature { NONE } -- Attribute(s)
	
	current_string : STRING

feature { NONE } -- Chars

	this_char(i : INTEGER) : CHARACTER is
		require
			ok: current_string /= Void and then (i >= 0 and i <= (current_string.count+1))
		do
			if i > 0 and i <= current_string.count then
				Result := current_string.item(i)
			end
		end

	next_char(i : INTEGER) : CHARACTER is
		require
			ok: current_string /= Void and then (i > 0 and i <= current_string.count)
		do
			if i < current_string.count then
				Result := current_string.item(i+1)
			end
		end

	previous_char(i : INTEGER) : CHARACTER is
		require
			ok: current_string /= Void and then (i > 0 and i <= current_string.count)
		do
			if i > 1 then
				Result := current_string.item(i-1)
			end
		end

feature -- Character type

	is_digit(c : CHARACTER) : BOOLEAN is
		do
			Result := c >= '0' and c <= '9'
		end

	is_punct(c : CHARACTER)	: BOOLEAN is
		do
			Result := c /= ' ' and not is_alpha(c) and not is_digit(c)
		end

	is_alpha(c : CHARACTER) : BOOLEAN is
		do
			Result := (c >= 'a' and c <= 'z') or (c >= 'A' and c <= 'Z')
		end

feature	-- Eiffel character type

	is_separator(c : CHARACTER) : BOOLEAN is
		do
			Result := c = ' ' or c = '%T' or is_punct(c)
		end

	is_identifier(c : CHARACTER) : BOOLEAN is
			-- Can this be the second (or more) character of an Eiffel identifier.
		do
			Result := is_alpha(c) or is_digit(c) or c = '_'
		end
	
	is_number(c : CHARACTER) : BOOLEAN is
			-- Can this be the second (or more) character of an Eiffel number?
		do
			Result := is_digit(c) or c = 'e' or c = 'E' or c = '.' or c = '_' 
		end

feature { NONE }
	
	process_line is
		require
			has_current: current_string /= Void
		local
			char : CHARACTER;
			i, j, count : INTEGER;
			output_done : BOOLEAN
		do
			from
				i := 1
				count := current_string.count
			variant
				count + 1 - i
			until
				i > count
			loop
				output_done := false
				char := this_char(i)

				if char = '-' and then next_char(i) = '-' then -- comment
					print_comment(current_string.substring(i,count))
					output_done := true
					i := count
				elseif (char = '%"' or char = '%'') and then i < count then -- quoted string
					from 
						j := i + 1
					invariant
						count_ok: j <= count
					variant
						count - j
					until
						j = count or else (this_char(j) = char and previous_char(j) /= '%%')					
					loop
						j := j + 1
					end

					print_quoted_string(current_string.substring(i,j))
					output_done := true
					i := j
				elseif is_digit(char) and then 
						is_separator(previous_char(i)) then -- number
					from
						j := i + 1
					variant
						count + 1 - j
					until
						j > count or else not is_number(this_char(j))
					loop
						j := j + 1
					end

					if this_char(j) = 'b' or this_char(j) = 'B' then
						j := j + 1
					end

					print_number(current_string.substring(i,j-1))
					output_done := true
					i := j - 1
				elseif is_alpha(char) and then -- or char = '_' and then 
						is_separator(previous_char(i)) then -- identifier, keyword
					from
						j := i + 1  
					variant
						count + 1 - j
					until
						j > count or else not is_identifier(this_char(j))
					loop
						j := j + 1
					end

					process_identifier(current_string.substring(i,j-1))
					output_done := true
					i := j - 1
				end

				if not output_done then print_character(char) end
				i := i + 1
			end
		end

feature { NONE }
	
	process_identifier(str : STRING) is
		require
			ok: str /= Void
		do
			if is_keyword(str) then
				print_keyword(str)
			else
				print_identifier(str)
			end
			-- could distinguish Constant, TYPE, normal identifier
		end

feature { PUBLIC_NONE }
	
	print_character(char : CHARACTER) is
		deferred
		end

	print_new_line is
		deferred
		end

feature { PUBLIC_NONE }
	
	print_keyword(str : STRING) is
		require
			ok: str	/= Void
		deferred
		end

	print_identifier(str : STRING) is
		require
			ok: str /= Void
		deferred
		end

	print_quoted_string(str : STRING) is
		require
			ok: str	/= Void
		deferred
		end

	print_number(str : STRING) is
		require
			ok: str	/= Void
		deferred
		end

	print_comment(str : STRING) is
		require
			ok: str	/= Void
		deferred
		end		

end -- class EIFFEL_PRINT
